"""
Constants
---------

Specific constants to be used in ``harpy``, to help with consistency.
"""
# Output Fileextensions --------------------------------------------------------

FILE_AMPS_EXT = ".amps{plane}"
FILE_FREQS_EXT = ".freqs{plane}"
FILE_LIN_EXT = ".lin{plane}"

# Column Names -----------------------------------------------------------------
# TODO use these everywhere (jdilly)

# Basic ---
COL_NAME = "NAME"

# Lin Files ---
COL_TUNE = "TUNE"
COL_AMP = "AMP"
COL_MU = "MU"

COL_NATTUNE = "NATTUNE"
COL_NATAMP = "NATAMP"
COL_NATMU = "NATMU"

COL_FREQ = "FREQ"
COL_PHASE = "PHASE"
COL_ERR = "ERR"
